
function HTML5_AudioDriver () {
  	if(arguments.callee._singletonInstance){
  	    return arguments.callee._singletonInstance;
  	}
  	arguments.callee._singletonInstance = this;

  	var self = this; self.name = "HTML5_AudioDriver"; // MANDATORY
    
    this.audioItems = [];
    this.outputDevices = [];
    
    this.incremental = 0;
    this.audioContext = null;
    
    this.init = function ( volume ) {
      if(typeof(volume) === "undefined"){volume = -1;}
      var AC = window.AudioContext||window.webkitAudioContext;
      self.audioContext = new AC();
      console.log("AUDIO# Init Audio Driver : "+self.name);
      self.defaultAudioOptions = {
        'channels': [{
          'channel': 0,
          'internal_id': -1,
          'gain': volume
        },{
          'channel': 1,
          'internal_id': -1,
          'gain': volume
        }],
        'deviceId' : "default",
        'muted' : false,
        'channel': -1,
        'gain':-1,
        'internal_id':-1
      };
      self.incremental = 0;
      self.audioItems = [];
      self.outputDevices = [];
      self.getOutputDevices();
    };
    
    this.resetDefaultAudioOptions = function(volume){
      if(typeof(volume) === "undefined"){volume = -1;}
      self.defaultAudioOptions = {
        'channels': [{
          'channel': 0,
          'internal_id': -1,
          'gain': volume
        },{
          'channel': 1,
          'internal_id': -1,
          'gain': volume
        }],
        'deviceId' : "default",
        'muted' : false,
        'channel': -1,
        'gain':-1,
        'internal_id':-1
      };
    };
    
    this.getChannelsFromDeviceId = function(options){
      var channels = [];
      self.defaultAudioOptions.channels.map(function(e){
        if(e.deviceId == options.deviceId && e.channel == options.channel){
          channels.push(e);
          return e.deviceId;
        }else{
          return false;
        }
      }).indexOf(options.deviceid);
      if(channels.length == 0){
        channels = [{
          'channel': 0,
          'internal_id': -1,
          'gain': self.defaultAudioOptions.gain
        },{
          'channel': 1,
          'internal_id': -1,
          'gain': self.defaultAudioOptions.gain
        }];
      }else if(channels.length == 1){
        if(channels[0].channel == 0){
          channels.push({
            'channel': 1,
            'internal_id': -1,
            'gain': self.defaultAudioOptions.gain
          });
        }else{
          channels.push({
            'channel': 0,
            'internal_id': -1,
            'gain': self.defaultAudioOptions.gain
          });
        }
      }
      return channels;
    };
    
    this.getAudioConfig = function(options){
      var configIndex = -1;
      try{
        var testConfigIndex = self.defaultAudioOptions.channels.map(function(e,index){
          if(typeof(options.channel) !== "undefined"){
            if(e.channel == options.channel && (e.internal_id == options.internal_id || e.internal_id == -1)){
              configIndex = index;
              return e.internal_id;
            }else{
              return false;
            }
          }else{
            if(e.internal_id == options.internal_id){
              configIndex = index;
              return e.internal_id;
            }else{
              return false;
            }
          }
        }).indexOf(options.internal_id);
      }catch(e){
        console.log("Audio Config for id "+options.internal_id+" not found for channel :"+options.channel);
      }
      return configIndex;
    };
    
    this.addAudioConfig = function(options){
      if(typeof(options) !== "undefined"){
        if(typeof(options.channel) !== "undefined" && typeof(options.internal_id) !== "undefined"){        
          var selectedChannel = self.getAudioConfig(options);
          if(selectedChannel<0){
            self.defaultAudioOptions.channels.push(options);
            console.log("Audio Config for id "+options.internal_id+" not found on defaultAudioOptions.channels ... creating on channel :"+options.channel);
          }else{
            self.defaultAudioOptions.channels[selectedChannel] = options;
          }
        }
      }
    };
    
    this.getDefaultAudioOptions = function(){
      return self.defaultAudioOptions;
    };
    
    this.setDefaultAudioOptions = function(options){
      if(typeof(options) !== "undefined"){  
        if(typeof(options.deviceId) !== "undefined"){
          if(options.deviceId != ""){
            if(self.defaultAudioOptions.deviceId != options.deviceId){
              self.defaultAudioOptions.deviceId = options.deviceId;
            }
          }
        }
        if(typeof(options.muted) !== "undefined"){
          if(self.defaultAudioOptions.muted != options.muted){
            self.defaultAudioOptions.muted = options.muted; 
          }
        } 
        
        
        var selectedChannel = self.getAudioConfig(options);
        if(selectedChannel>=0){
          if(typeof(options.internal_id) !== "undefined"){
            if(options.internal_id != ""){
              if(self.defaultAudioOptions.channels[selectedChannel].internal_id != options.internal_id){
                self.defaultAudioOptions.channels[selectedChannel].internal_id = options.internal_id;
              }
              self.defaultAudioOptions.internal_id = options.internal_id;
            }
          }
          if(typeof(options.channel) !== "undefined"){
            if(self.defaultAudioOptions.channels[selectedChannel].channel != options.channel){
              self.defaultAudioOptions.channels[selectedChannel].channel = options.channel; 
            }
            self.defaultAudioOptions.channel = options.channel; 
          } 
          if(typeof(options.gain) !== "undefined"){
            if(self.defaultAudioOptions.channels[selectedChannel].gain != options.gain){
              if(options.gain >= 0){
                self.defaultAudioOptions.channels[selectedChannel].gain = options.gain;
                self.defaultAudioOptions.gain = options.gain;
              }else{
                self.defaultAudioOptions.channels[selectedChannel].gain = 1.0;
                self.defaultAudioOptions.gain = options.gain;
              }
            }else{
              if(options.gain >= 0){
                self.defaultAudioOptions.gain = options.gain;
              }else{
                self.defaultAudioOptions.gain = options.gain;
              }
            }
          }
        }
      }
    };
    
    this.setAudioOptions = function(options,audioObj){
      var setting_default = true;
      if(typeof(audioObj) !== "undefined"){setting_default = false;}
      var changed = false;
      if(typeof(options) !== "undefined"){
        if(typeof(options.deviceId) !== "undefined"){
          if(options.deviceId != ""){
            if(!setting_default){
              if(audioObj.options.deviceId != options.deviceId){
                audioObj.options.deviceId = options.deviceId;
                changed = true;
              }
            }else{
              if(self.defaultAudioOptions.deviceId != options.deviceId){
                self.defaultAudioOptions.deviceId = options.deviceId;
                changed = true;
              }
            }
          }
        }
        if(typeof(options.muted) !== "undefined"){
          if(!setting_default){
            if(audioObj.options.muted != options.muted){
              audioObj.options.muted = options.muted; 
              changed = true;
            }
          }else{
            if(self.defaultAudioOptions.muted != options.muted){
              self.defaultAudioOptions.muted = options.muted; 
              changed = true;
            }
          }
        } 
        if(!setting_default){
          if(typeof(options.internal_id) !== "undefined"){
            if(options.internal_id != ""){
              if(audioObj.options.internal_id != options.internal_id){
                audioObj.options.internal_id = options.internal_id;
                changed = true;
              }
            }
          }
          
          if(typeof(options.channel) !== "undefined"){
            if(audioObj.options.channel != options.channel){
              audioObj.options.channel = options.channel; 
              changed = true;
            }
          } 
          
          if(typeof(options.gain) !== "undefined"){
            if(audioObj.options.gain != options.gain){
              if(options.gain >= 0){
                audioObj.options.gain = options.gain;
              }else{
                audioObj.options.gain = 1.0;
              }
              changed = true;
            }
          }
        }else{
        } 
        
        if(!setting_default){
          console.log("Audio Object Changed ["+changed+"] Set Up  with options :");
          console.log(audioObj.options);     
        }else{
          console.log("Default Audio Settings Changed ["+changed+"] Set Up  options :"); 
          console.log(self.defaultAudioOptions);    
        }  
      }      
      return changed;
    };
    

    
    this.refreshAudioOutput = function(audioObj,options,source){
      var changes = self.setAudioOptions(options,audioObj);
      var itemIndex = self.audioItems.map(function(e){return e.item_number;}).indexOf(audioObj.item_number);

      var channels = 0;
      var i=0;
      if(itemIndex >= 0){
        if(changes){
          console.log("Updating item : "+audioObj.item_number+" on Channel index : "+itemIndex);
          var real_device_id = 0;
          if(audioObj.options.channel >= 0){
            real_device_id = self.getRealDeviceId(self.outputDevices,audioObj.options.deviceId);
            if(real_device_id){
              if(self.audioItems[itemIndex].deviceId !== real_device_id){
                self.audioItems[itemIndex].audio.setSinkId(real_device_id).then(function(e){
                  console.log("Updating Audio SinkID OK");
                }).catch(function(e){
                  console.log("Updating Audio SinkID BAD");
                }); 
              }
            }
            if(self.audioItems[itemIndex].audio.muted != audioObj.options.muted){
              self.audioItems[itemIndex].audio.muted = audioObj.options.muted;
            }
            
            if(self.audioItems[itemIndex].options.channel >= 0){
              self.audioItems[itemIndex].gains[self.audioItems[itemIndex].options.channel].gain.value = audioObj.options.gain;
            }else{
              channels = self.getChannelsFromDeviceId(self.audioItems[itemIndex].options);
              for(i=0;i<channels.length;i++){
                self.audioItems[itemIndex].gains[i].gain.value = audioObj.options.gain;
              }
            }
          }else{
            real_device_id = self.getRealDeviceId(self.outputDevices,audioObj.options.deviceId);
            if(real_device_id){
              if(self.audioItems[itemIndex].deviceId !== real_device_id){
                self.audioItems[itemIndex].audio.setSinkId(real_device_id).then(function(e){
                  console.log("Updating Audio SinkID OK");
                }).catch(function(e){
                  console.log("Updating Audio SinkID BAD");
                }); 
              }
            }
            if(self.audioItems[itemIndex].audio.muted != audioObj.options.muted){
              self.audioItems[itemIndex].audio.muted = audioObj.options.muted;
            }
            self.audioItems[itemIndex].audio.volume = audioObj.options.gain;
          }
        }
        if(typeof(source) !== "undefined"){
          console.log("Updating Output for channel source : "+source+" on deviceid : "+ audioObj.options.deviceId);
          if(audioObj.options.channel >= 0){
            try{
              if(typeof(self.audioItems[itemIndex].audio_source) !== "undefined"){
                self.audioItems[itemIndex].audio_source.disconnect();
                self.audioItems[itemIndex].audio.srcObject = null;
              }
              if(typeof(self.audioItems[itemIndex].media_stream) !== "undefined"){
                self.audioItems[itemIndex].media_stream.disconnect();
                self.audioItems[itemIndex].media_stream.stream = undefined;
              }
              self.audioItems[itemIndex].media_stream = undefined;

              self.audioItems[itemIndex].audio_source = undefined;
              self.audioItems[itemIndex].audio_source = self.audioContext.createMediaElementSource(source);
              if(typeof(self.audioItems[itemIndex].splitter) !== "undefined"){self.audioItems[itemIndex].splitter.disconnect();}
              self.audioItems[itemIndex].audio_source.connect(self.audioItems[itemIndex].splitter);
              
              channels = self.getChannelsFromDeviceId(self.audioItems[itemIndex].options);
              for(i=0;i<channels.length;i++){
                if(typeof(self.audioItems[itemIndex].gains[i]) !== "undefined"){
                  self.audioItems[itemIndex].gains[i].disconnect();
                }else{
                  self.audioItems[itemIndex].gains[i] = self.audioContext.createGain();
                }
                if(audioObj.options.channel >= 0){
                  self.audioItems[itemIndex].gains[i].gain.value = 0.0;
                }else{
                  self.audioItems[itemIndex].gains[i].gain.value = 1.0;
                }
                self.audioItems[itemIndex].splitter.connect(self.audioItems[itemIndex].gains[i], 0, 0);
                self.audioItems[itemIndex].gains[i].connect(self.audioItems[itemIndex].merger, 0, i);
              }
              if(audioObj.options.channel >= 0){
                self.audioItems[itemIndex].gains[audioObj.options.channel].gain.value = audioObj.options.gain;
              }
              self.audioItems[itemIndex].media_stream = self.audioContext.createMediaStreamDestination();
              self.audioItems[itemIndex].audio.srcObject = self.audioItems[itemIndex].media_stream.stream;
              if(typeof(self.audioItems[itemIndex].merger) !== "undefined"){self.audioItems[itemIndex].merger.disconnect();}
              self.audioItems[itemIndex].merger.connect(self.audioItems[itemIndex].media_stream);
            }catch(e){
              console.log("AUDIO# Failed to UPDATE AudioContext connections: ");
              console.log(e);
            }
          }else{
            self.audioItems[itemIndex].audio = source;
          }
        }
      }
    };
    
    this.createAudioObject = function(){
      var options = {
        'deviceId' : self.defaultAudioOptions.deviceId,
        'muted' : self.defaultAudioOptions.muted,
        'channel': self.defaultAudioOptions.channel,
        'gain': self.defaultAudioOptions.gain,
        'internal_id': self.defaultAudioOptions.internal_id
    };
      
      return {item_number : self.incremental, options : options};
    };
    
    this.connectAudioOutput = function (source,options) {
      var audioObj = self.createAudioObject();
      self.setAudioOptions(options,audioObj);
      var real_device_id = self.getRealDeviceId(self.outputDevices,audioObj.options.deviceId);  
      if(real_device_id){
        self.audioItems.push({
          audio:undefined,
          splitter : undefined,
          merger : undefined,
          gains : [],
          audio_source : undefined,
          media_stream : undefined,
          options :audioObj.options,
          item_number : self.incremental,
          fade_timer : 0
        });
        self.incremental++;
        var itemIndex = self.audioItems.map(function(e){return e.item_number;}).indexOf(audioObj.item_number);
        console.log("Creating item : "+audioObj.item_number+" on Channel index : "+itemIndex);
        if(itemIndex >= 0){
          
          if(audioObj.options.channel >= 0){
            var channels = self.getChannelsFromDeviceId(self.audioItems[itemIndex].options);
            
            if(typeof(self.audioItems[itemIndex].audio) === "undefined"){
              self.audioItems[itemIndex].audio = new Audio();
              self.audioItems[itemIndex].audio.autoplay = true;
              self.audioItems[itemIndex].splitter =  self.audioContext.createChannelSplitter(channels.length);
              self.audioItems[itemIndex].merger = self.audioContext.createChannelMerger(channels.length);
            }else{
              self.audioItems[itemIndex].audio.srcObject = null;
              self.audioItems[itemIndex].merger.disconnect();
              self.audioItems[itemIndex].splitter.disconnect();
            }
            self.audioItems[itemIndex].audio.muted = audioObj.options.muted;
            self.audioItems[itemIndex].media_stream = undefined;

            self.audioItems[itemIndex].audio_source = undefined;
            self.audioItems[itemIndex].audio.setSinkId(real_device_id).then(function(e){
              console.log("Creating Audio SinkID OK");
              try{
                self.audioItems[itemIndex].audio_source = self.audioContext.createMediaElementSource(source);
                self.audioItems[itemIndex].audio_source.connect(self.audioItems[itemIndex].splitter);
                for(var i=0;i<channels.length;i++){
                  if(typeof(self.audioItems[itemIndex].gains[i]) === "undefined"){
                    self.audioItems[itemIndex].gains[i] = self.audioContext.createGain();
                  }else{
                    self.audioItems[itemIndex].gains[i].disconnect();
                  }
                  if(audioObj.options.channel >= 0){
                    self.audioItems[itemIndex].gains[i].gain.value = 0.0;
                  }else{
                    self.audioItems[itemIndex].gains[i].gain.value = 1.0;
                  }
                  self.audioItems[itemIndex].splitter.connect(self.audioItems[itemIndex].gains[i], 0, 0);
                  self.audioItems[itemIndex].gains[i].connect(self.audioItems[itemIndex].merger, 0, i);
                }
                if(audioObj.options.channel >= 0){
                  self.audioItems[itemIndex].gains[audioObj.options.channel].gain.value = audioObj.options.gain;
                }
                self.audioItems[itemIndex].media_stream = self.audioContext.createMediaStreamDestination();
                self.audioItems[itemIndex].audio.srcObject = self.audioItems[itemIndex].media_stream.stream;
                self.audioItems[itemIndex].merger.connect(self.audioItems[itemIndex].media_stream);
              }catch(ex){
                console.log("AUDIO# Failed to CREATE AudioContext :");
                console.log(ex);
              }
            }).catch(function(e){
              console.log("Creating Audio SinkID BAD");
            });
          }else{
            self.audioItems[itemIndex].audio = source;
            self.audioItems[itemIndex].audio.setSinkId(real_device_id).then(function(e){
              console.log("Creating Audio SinkID OK");
              try{
                self.audioItems[itemIndex].audio.volume = audioObj.options.gain;
              }catch(ex){
                console.log("AUDIO# Failed to CREATE AudioContext :");
                console.log(ex);
              }
            }).catch(function(e){
              console.log("Creating Audio SinkID BAD");
            });
          }
        }
      }else{
          console.log("AUDIO# No Real DeviceId found from value : "+audioObj.options.deviceId);
      }
      return audioObj;
      
    };
    
    this.fadeAudioOutput = function(audioObj,destOptions,duration){
      console.log("AUDIO# Trying to do fade operation");
      try{
        var itemIndex = self.audioItems.map(function(e){return e.item_number;}).indexOf(audioObj.item_number);
        if(itemIndex>=0){
          if(destOptions.gain != audioObj.options.gain){
            if(self.audioItems[itemIndex].fade_timer > 0){
              console.log("Force clear fade timer : "+self.audioItems[itemIndex].fade_timer);
              clearTimeout(self.audioItems[itemIndex].fade_timer);
              self.audioItems[itemIndex].fade_timer  = 0;
            }
            console.log("AUDIO# Fade Starting ...");
            var step = (destOptions.gain - audioObj.options.gain) / (duration / 100);  
            step = Math.round(step * 10000) / 10000; 
            var fade_status = {
              item_number : audioObj.item_number,
              gain : audioObj.options.gain,
              dest_gain : destOptions.gain,
              step : step,
              remaining : duration,
              end_muted : false,
              index: itemIndex
            };
            if(typeof(destOptions.muted) !== "undefined" && destOptions.gain == 0.0){
              fade_status.end_muted = destOptions.muted;
            }
            console.log("AUDIO# Fade Initial Status:");
            self.audioItems[itemIndex].fade_timer = setTimeout(self.processFadeOperation.bind(null,fade_status),100);
          }else{
            self.refreshAudioOutput(self.getAudioObjectByItemNumber(self.audioItems[itemIndex].item_number),destOptions); 
          }
        }else{
          console.log("AUDIO# Fade no Item found");
        }
      }catch(e){
        console.log("AUDIO# Fade Exception ...");
        console.log(e);
      } 
    };
    
    this.processFadeOperation = function(status){
      try{
        if(typeof(self.audioItems[status.index]) !== "undefined"){
          clearTimeout(self.audioItems[status.index].fade_timer);
          self.audioItems[status.index].fade_timer = 0;
          if(status.gain != status.dest_gain && status.remaining > 0){
            status.remaining-=100;
            var compare = status.gain+status.step;
            compare = Math.round(compare * 10000) / 10000;
            if(compare >= 0.0 && compare <= 1.0){
              status.gain+=status.step;
              status.gain = Math.round(status.gain * 10000) / 10000; 
              self.refreshAudioOutput(self.getAudioObjectByItemNumber(status.item_number),{'channel':self.audioItems[status.index].options.channel,'gain':status.gain});   
            }
            self.audioItems[status.index].fade_timer = setTimeout(self.processFadeOperation.bind(null, status),100);
          }else{
            self.refreshAudioOutput(self.getAudioObjectByItemNumber(status.item_number),{'gain':status.dest_gain,'muted':status.end_muted});  
            
          }
        }
      }catch(e){
        console.log("Something went wrong while fading ... ");
        console.log(e);
        try{
          console.log("Status should be : ");
          console.log(status);
        }catch(e){
          console.log(e);
        }
      }
    };
    
    this.disconnectAudioOutput = function(audioObj){
      var busy = false;
      if(audioObj){
        busy = true;
        var itemIndex = self.audioItems.map(function(e){return e.item_number;}).indexOf(audioObj.item_number);
        console.log("Removing item : "+audioObj.item_number+" on Channel index : "+itemIndex);
        if(itemIndex >= 0){
          if(typeof(self.audioItems[itemIndex]) !== "undefined"){
            try{
              if(self.audioItems[itemIndex].fade_timer>0){
                clearTimeout(self.audioItems[itemIndex].fade_timer);
              }
            }catch(e){
              console.log(e);
            }
            console.log("AUDIO# DisconnectAudioOutput ...");
            console.log(audioObj);
            console.log(self.audioItems[itemIndex]);
            try{
              if(typeof(self.audioItems[itemIndex].audio) !== "undefined"){
                self.audioItems[itemIndex].audio.srcObject = null;
                self.audioItems[itemIndex].audio = null;
                self.audioItems[itemIndex].audio = undefined;
              }
              
              if(typeof(self.audioItems[itemIndex].splitter) !== "undefined"){
                self.audioItems[itemIndex].splitter.disconnect();
                self.audioItems[itemIndex].splitter = null;
              }
              if(typeof(self.audioItems[itemIndex].audio_source) !== "undefined"){
                self.audioItems[itemIndex].audio_source.disconnect();
              }
              if(typeof(self.audioItems[itemIndex].media_stream) !== "undefined"){
                self.audioItems[itemIndex].media_stream.disconnect();
                self.audioItems[itemIndex].media_stream.stream = undefined;
              }
              self.audioItems[itemIndex].media_stream = undefined;

              self.audioItems[itemIndex].audio_source = undefined;          
              var channels = self.getChannelsFromDeviceId(self.audioItems[itemIndex].options);
              for(var i=0;i<channels.length;i++){
                if(typeof(self.audioItems[itemIndex].gains[i]) !== "undefined"){
                  self.audioItems[itemIndex].gains[i].disconnect();
                  self.audioItems[itemIndex].gains[i] = null;
                }
              }
              if(typeof(self.audioItems[itemIndex].merger) !== "undefined"){
                self.audioItems[itemIndex].merger.disconnect();
                self.audioItems[itemIndex].merger = null;
              }
              self.audioItems.splice(itemIndex,1);
              busy = false;
            }catch(e){
              console.log("AUDIO# Failed to Delete AudioContext connections: ");
              console.log(e);
            }      
          }  
        }
        
      }
      return busy;
    };
    
    this.getAudioObjectsByOptions = function(options){
      var items = [];
      var itemIndex = self.audioItems.map(function(e){
        if(e.options.channel == options.channel && e.options.deviceId == options.deviceId){
          items.push({'item_number':e.item_number,'options':e.options});
          return e.options;
        }else{
          return -1;
        }
      }).indexOf(options);
      return items;
    };

    this.getAudioObjectsByChannel = function(channel){
      var items = [];
      var itemIndex = self.audioItems.map(function(e){
        if(e.options.channel == channel){
          items.push({'item_number':e.item_number,'options':e.options});
          return e.channel;
        }else{
          return -1;
        }
      }).indexOf(channel);
      return items;
    };
    
    this.getAudioObjectsByDeviceId = function(deviceId){
      var items = [];
      var itemIndex = self.audioItems.map(function(e){
        if(e.options.deviceId == deviceId){
          items.push({'item_number':e.item_number,'options':e.options});
          return e.deviceId;
        }else{
          return -1;
        }
      }).indexOf(deviceId);
      return items;
    };
    
    this.getAudioObjectsByInternalId = function(internal_id){
      var items = [];
      var itemIndex = self.audioItems.map(function(e){
        if(e.options.internal_id == internal_id){
          items.push({'item_number':e.item_number,'options':e.options});
          return e.item_number;
        }else{
          return -1;
        }
      }).indexOf(internal_id);
      return items;
    };
    
    this.getAudioObjectByItemNumber = function(item_number){
      var itemIndex = self.audioItems.map(function(e){return e.item_number;}).indexOf(item_number);
      if(itemIndex >= 0){
        return {'item_number':item_number,'options':self.audioItems[itemIndex].options};
      }else{
        return false;
      }
    };
    
    this.getRealDeviceId = function(devices,value){
      var real_device_id = false;
      var index = -1;
      if(isNaN(value)){
        index = devices.map(function(e){return e.label;}).indexOf(value);
      }else{
        index = devices.map(function(e){return e.groupId;}).indexOf(value);
      }
      
      if(index >= 0){
        real_device_id =devices[index].deviceId;
      }
      return real_device_id;
    };
    
    this.getOutputDevices = function(callback){
      try{
        navigator.mediaDevices.enumerateDevices().then(function(devices) {
          self.outputDevices = [];
          for(var i=0;i < devices.length;i++){
            if(devices[i].kind == "audiooutput"){
              self.outputDevices.push(devices[i]);
            }
          }
          if(typeof(callback) !== "undefined"){
            callback(self.outputDevices);
          }
        }).catch(function(err) {
          console.log(err.name + ": " + err.message);
        });
      }catch(e){
        console.log(e);
      }
    };
    
    this.getAllDevices = function(callback){
      try{
        navigator.mediaDevices.enumerateDevices().then(function(devices) {
          if(typeof(callback) !== "undefined"){
            callback(devices);
          }
        }).catch(function(err) {
          console.log(err.name + ": " + err.message);
        });
      }catch(e){
        console.log(e);
      }
    };
    
    this.unload = function(){
      try{
        for(var i=0;i<self.audioItems.length;i++){
          self.disconnectAudioOutput(self.audioItems[i]);
        }
        self.audioItems = null;
        self.outputDevices = null;
        self.defaultAudioOptions = null;
      }catch(e){
        console.log(e);
      }
      
    };
}
